from ._version import version

__version__ = version

from clouddrift.raggedarray import RaggedArray
import clouddrift.adapters
import clouddrift.datasets
import clouddrift.kinematics
import clouddrift.pairs
import clouddrift.plotting
import clouddrift.ragged
import clouddrift.signal
import clouddrift.sphere
import clouddrift.wavelet
